import os
import codecs
import tempfile
import urllib.parse
import hashlib
import subprocess
import logging
import shutil
from collections import namedtuple, OrderedDict

from google.cloud.storage.client import Client
from google.api_core.exceptions import NotFound

import portalocker

logger = logging.getLogger(__name__)

DEFAULT_GCP_PROJECT = 'freenome-computational'
DEFAULT_LOCAL_PREFIX = '/srv/reference_data/'
DEFAULT_REMOTE_PREFIX = 'gs://balrog/reference-data/'


class FileAlreadyExistsError(Exception):
    pass


class FileMismatchError(Exception):
    pass


class MissingFileError(Exception):
    pass


class KeyAlreadyExistsError(Exception):
    pass


def hex_to_base64(hex_str):
    return codecs.encode(codecs.decode(hex_str, 'hex'), 'base64').strip().decode('ascii')


def calc_md5sum_from_fname(fname):
    hex_str = subprocess.run(
        ["md5sum", fname],
        stdout=subprocess.PIPE
    ).stdout.split()[0].decode('ascii')
    return hex_to_base64(hex_str)


def calc_md5sum_from_fp(fp):
    fpos = fp.tell()
    m = hashlib.md5()
    fp.seek(0)
    m.update(fp.read().encode('utf8'))
    fp.seek(fpos)
    return hex_to_base64(m.hexdigest())


def _get_gcs_blob(remote_prefix, remote_relative_path):
    absolute_remote_path = remote_prefix + remote_relative_path
    res = urllib.parse.urlsplit(absolute_remote_path)
    rel_path = res.path[1:]
    blob = Client(DEFAULT_GCP_PROJECT).bucket(res.netloc).blob(rel_path)
    return blob


DataManifestRecord = namedtuple(
    'DataManifestRecord',
    ['name', 'local_path', 'remote_path', 'md5sum', 'size', 'notes']
)


class DataManifest(OrderedDict):
    def _get_gcs_blob(self, remote_relative_path):
        return _get_gcs_blob(self.remote_prefix, remote_relative_path)

    def _save_to_disk(self):
        """Save the current data to disk."""
        with portalocker.Lock(self.fname, "r+") as fp:
            # first make sure that the manifest hasn't changed since we last read it
            on_disk_md5sum = calc_md5sum_from_fp(fp)
            if on_disk_md5sum != self._md5sum:
                raise RuntimeError(
                    f"'{self.fname}' was modified by another program (current md5sum "
                    f"'{on_disk_md5sum}' vs '{self._md5sum}')"
                )

            # truncate the file, and re-write it
            fp.seek(0)
            fp.truncate()
            fp.write("\t".join(self.header) + "\n")
            for record in self.values():
                fp.write("\t".join(record) + "\n")
            fp.flush()

            # update the md5sum
            self._md5sum = calc_md5sum_from_fp(fp)
            fp.close()

    def __init__(
            self,
            manifest_fname,
            remote_prefix=DEFAULT_REMOTE_PREFIX
    ):
        self.fname = manifest_fname
        self.remote_prefix = remote_prefix

        self.header = None

        # we lock the file to ensure that only one file is doing this at a time
        with portalocker.Lock(self.fname, 'r') as fp:
            # store the md5sum so that we can tell
            # if the file has been modified before writing it out
            self._md5sum = calc_md5sum_from_fp(fp)

            # read all of the file contents into memory
            for line_i, line in enumerate(fp):
                # read the header
                if line_i == 0:
                    self.header = line.strip("\n").split("\t")
                    continue
                # skip empty lines
                if line.strip() == '':
                    continue
                # parse and store this record to the ordered dict
                record = DataManifestRecord(*line.strip("\n").split("\t"))
                if record.name in self:
                    raise KeyAlreadyExistsError("'{record.name}' is duplicated in '{self.fname}'")
                self[record.name] = record

    def remove_file(self, name):
        """Remove a file from the manifest.

        This does *not* remove the file from the filesystem or from GCS.
        """
        del self[name]
        self._save_to_disk()

    def add_file(
            self,
            name,
            fname,
            local_relative_path,
            remote_relative_path,
            note=''
    ):
        """Add a file to the manifest.

        Add a file to the manifest and upload the file to GCS.
        """
        if name in self:
            raise KeyAlreadyExistsError("'{record.name}' is duplicated in '{self.fname}'")

        # make sure that we can open the file that we want to add for reading
        with open(fname) as _: # noqa
            pass

        # find the file's file size and calculate the checksum
        logger.info(f"Calculating md5sum for '{fname}'")
        local_md5sum = calc_md5sum_from_fname(fname)
        logger.debug(f"Calculated md5sum '{local_md5sum}' for '{fname}'.")
        local_fsize = os.path.getsize(fname)
        logger.debug(f"Calculated filesize '{local_fsize}' for '{fname}'.")

        # setup the remote file object
        blob = self._get_gcs_blob(remote_relative_path)
        try:
            blob.reload()
            # if it exists, make sure that it is the same as the local file
            if local_md5sum != blob.md5_hash:
                raise FileAlreadyExistsError(
                    f"File '{self.remote_prefix}{remote_relative_path}' already exists with md5sum"
                    f"'{blob.md5_hash}' vs '{local_md5sum}' for '{fname}')"
                )
            if local_fsize != blob.size:
                raise FileAlreadyExistsError(
                    f"File '{self.remote_prefix}{remote_relative_path}' already exists with file "
                    f"size '{blob.size}' vs '{local_fsize}' for '{fname}')"
                )
        # if we can't find the file, upload it
        except NotFound:
            logger.info(f"Uploading '{fname}' to '{self.remote_prefix}{remote_relative_path}'")
            blob.upload_from_filename(fname)
            assert blob.size == local_fsize, \
                "We just uploaded this file so the filesizes should match"
            assert blob.md5_hash == local_md5sum, \
                f"We just uploaded this file so the md5sums should match " \
                f"('{blob.md5_hash}' vs '{local_md5sum}')"
        assert blob.md5_hash is not None
        assert blob.size is not None

        self[name] = DataManifestRecord(
            name, local_relative_path, remote_relative_path, local_md5sum, str(local_fsize), note
        )
        self._save_to_disk()

    def _verify_record(self, record, local_abs_path, check_md5sums=True):
        """Verify that the file at 'local_abs_path' matches that in record.

        If check_md5sums is True then verify that the md5sums match (this is slow).
        """
        # check that the file exists
        if not os.path.exists(local_abs_path):
            raise MissingFileError("Can not find '{record.name}' at '{local_abs_path}'")

        # ensure the filesizes match
        local_fsize = os.path.getsize(local_abs_path)
        logger.debug(f"Calculated filesize '{local_fsize}' for '{local_abs_path}'.")
        if local_fsize != int(record.size):
            raise FileMismatchError(
                f"'{local_abs_path}' has size '{local_fsize}' vs '{record.size}' in the manifest")

        # ensure the md5sum matches
        if check_md5sums:
            logger.info(f"Calculating md5sum for '{local_abs_path}'.")
            local_md5sum = calc_md5sum_from_fname(local_abs_path)
            logger.debug(f"Calculated md5sum '{local_md5sum}' for '{local_abs_path}'.")
            if local_md5sum != record.md5sum:
                raise FileMismatchError(
                    f"'{local_abs_path}' has md5sum '{local_md5sum}' "
                    f"vs '{record.md5sum}' in the manifest"
                )

    def _sync_record(self, record, local_abs_path):
        # if local_path already exists, then make sure that it matches the remote file
        if os.path.exists(local_abs_path):
            self._verify_record(record, local_abs_path)
        # otherwise, copy it to the correct location
        else:
            # copy the file to local_path
            logger.info(f"Copying '{record.remote_path}' to '{local_abs_path}'.")
            blob = self._get_gcs_blob(record.remote_path)
            # make sure the directory exists
            dirname = os.path.dirname(local_abs_path)
            if not os.path.exists(dirname):
                os.makedirs(dirname)
            blob.download_to_filename(local_abs_path)
            assert os.path.getsize(local_abs_path) == int(record.size), "{}, {}".format(
                os.path.getsize(local_abs_path), record.size
            )
            # skip this assert because it is slow (and filesize should catch anything weird)
            # assert calc_md5sum(local_path) == local_md5sum

    def sync(self, local_prefix):
        """Sync the remote files to a local path."""
        for record in self.values():
            local_abs_path = os.path.join(local_prefix, record.local_path)
            self._sync_record(record, local_abs_path)

    def verify(self, local_prefix, check_md5sums=False):
        """Ensure that the files at 'local_prefix' match the manifest.

        If 'slow' is True, then additionally ensure that the md5sum's match.
        """
        for record in self.values():
            local_abs_path = os.path.join(local_prefix, record.local_path)
            self._verify_record(record, local_abs_path, check_md5sums=check_md5sums)


def _add_base_to_path(rel_path):
    return os.path.normpath(os.path.join(os.path.abspath(os.path.dirname(__file__)), rel_path))


LOCAL_TEST_DATA_PREFIX = "PLACEHOLDER FOR FLAKE8 BUT NEEDS TO BE FIXED"

TEST_MANIFEST_FNAME = _add_base_to_path('../tests/data/pipeline-data/data-manifest.tsv')
TEST_DATA_FILE = _add_base_to_path('../tests/data/pipeline-data/eight_As.fa')

TEST_LOCAL_RELATIVE_PATH = 'eight_As.fa'
TEST_REMOTE_RELATIVE_PATH = 'eight_As.fa'
TEST_REMOTE_PATH = DEFAULT_REMOTE_PREFIX + TEST_REMOTE_RELATIVE_PATH

TEST_DATA_FILE_2 = os.path.normpath(os.path.join(
    os.path.abspath(os.path.dirname(__file__)),
    '../tests/data/pipeline-data/hg38.chrom.interval.chrM.bed.gz')
)
TEST_2_LOCAL_RELATIVE_PATH = './hg38/hg38.chrom.interval.chrM.bed.gz'


def reset_test_manifest():
    shutil.copy(TEST_MANIFEST_FNAME+".orig", TEST_MANIFEST_FNAME)


def add_file_to_manifest_and_verify():
    """Test that adding a file to the manifest works.

    This is called by the more specific test below.
    """
    # TODO - revert these to static global variables
    manifest_fname = TEST_MANIFEST_FNAME
    remote_prefix = DEFAULT_REMOTE_PREFIX
    test_data_fname = TEST_DATA_FILE
    local_relative_path = TEST_LOCAL_RELATIVE_PATH
    remote_relative_path = TEST_REMOTE_RELATIVE_PATH
    try:
        print(f"Loading manifest: {manifest_fname}")
        manifest = DataManifest(manifest_fname, remote_prefix=remote_prefix)
        manifest.add_file('eight_As', test_data_fname, local_relative_path, remote_relative_path)

        # read the local data file into a string
        # we use this below to ensure that it exists
        print(f"Reading local file: {test_data_fname}")
        with open(test_data_fname, 'rb') as ifp:
            local_data = ifp.read()

        # check that the file exists in GCS and is the same as the local file
        print(f"Downloading remote file: {DEFAULT_REMOTE_PREFIX}{TEST_REMOTE_RELATIVE_PATH}")
        blob = _get_gcs_blob(DEFAULT_REMOTE_PREFIX, TEST_REMOTE_RELATIVE_PATH)
        gcs_data = blob.download_as_string()
        assert local_data == gcs_data, f"local data: {local_data}\ngcs data: {gcs_data}"
    finally:
        print(f"Reseting manifest: {manifest_fname}")
        reset_test_manifest()
        print(f"Deleting remote file ({DEFAULT_REMOTE_PREFIX}{TEST_REMOTE_RELATIVE_PATH})")
        blob = _get_gcs_blob(DEFAULT_REMOTE_PREFIX, TEST_REMOTE_RELATIVE_PATH)
        blob.delete()


def test_add_duplicate_key():
    # test that we get an error if we try to add this file again with the same key
    try:
        manifest = DataManifest(TEST_MANIFEST_FNAME, DEFAULT_REMOTE_PREFIX)
        manifest.add_file(
            'eight_As', TEST_DATA_FILE, TEST_LOCAL_RELATIVE_PATH, TEST_REMOTE_RELATIVE_PATH)
        manifest.add_file(
            'eight_As', TEST_DATA_FILE, TEST_LOCAL_RELATIVE_PATH, TEST_REMOTE_RELATIVE_PATH)
    # this is what we expect. Any other exception will be propogated.
    except KeyAlreadyExistsError:
        pass

    # no expection is enexpected
    else:
        assert False, "Expected to see a 'KeyAlreadyExistsError'"
    finally:
        reset_test_manifest()


def test_add_remote_missing_file():
    """Test adding a file that is not in GCS."""
    # this is a vanilla add_file_to_manifest_and_verify call
    add_file_to_manifest_and_verify()


def test_add_remote_present_file():
    """Test adding a file that does not exist in gcs."""
    # add the file and then ensure that adding to the manifest still works
    blob = _get_gcs_blob(DEFAULT_REMOTE_PREFIX, TEST_REMOTE_RELATIVE_PATH)
    blob.upload_from_filename(TEST_DATA_FILE)
    add_file_to_manifest_and_verify()


def test_add_remote_mismatch_file():
    """Test adding a file that exists in gcs but does not match the file being added."""
    # add a mismatched file and ensure that we get an error
    blob = _get_gcs_blob(DEFAULT_REMOTE_PREFIX, TEST_REMOTE_RELATIVE_PATH)
    blob.upload_from_string("NATHAN")
    try:
        add_file_to_manifest_and_verify()
    except FileAlreadyExistsError:
        # we should get a FileAlreadyExistsError
        pass
    else:
        assert False, "This should have failed due to a file mismatch."


def test_locking_works():
    """Make sure that we get an error if two people try to modify the same process."""
    manifest_fname = TEST_MANIFEST_FNAME
    test_data_fname = TEST_DATA_FILE
    remote_prefix = DEFAULT_REMOTE_PREFIX
    remote_relative_path = TEST_REMOTE_RELATIVE_PATH
    local_relative_path = TEST_LOCAL_RELATIVE_PATH
    try:
        print(f"Loading manifest 1: {manifest_fname}")
        manifest_1 = DataManifest(manifest_fname, remote_prefix=remote_prefix)
        print(f"Loading manifest 2: {manifest_fname}")
        manifest_2 = DataManifest(manifest_fname, remote_prefix=remote_prefix)
        # this should succeed
        print(f"Adding file to manifest 1.")
        manifest_1.add_file('eight_As', test_data_fname, local_relative_path, remote_relative_path)
        # this should fail
        print(f"Adding file to manifest 2.")
        manifest_2.add_file('eight_As', test_data_fname, local_relative_path, remote_relative_path)
    except RuntimeError as inst:
        print(f"Adding file to manifest 2 FAILED with {inst}.")
        pass
    finally:
        print(f"Reseting manifest: {manifest_fname}")
        reset_test_manifest()
        print(f"Deleting remote file ({remote_relative_path})")
        _get_gcs_blob(DEFAULT_REMOTE_PREFIX, TEST_REMOTE_RELATIVE_PATH).delete()


def test_adding_two_files_works():
    try:
        test_data_fname = TEST_DATA_FILE
        remote_relative_path = TEST_REMOTE_RELATIVE_PATH
        local_relative_path = TEST_LOCAL_RELATIVE_PATH
        print(f"Loading manifest: {TEST_MANIFEST_FNAME}")
        manifest = DataManifest(TEST_MANIFEST_FNAME, remote_prefix=DEFAULT_REMOTE_PREFIX)
        print(f"Adding file 1 to manifest.")
        manifest.add_file('eight_As', test_data_fname, local_relative_path, remote_relative_path)
        print(f"Adding file 2 to manifest.")
        manifest.add_file(
            'eight_As_v2', test_data_fname, local_relative_path+'.v2', remote_relative_path+'.v2')
    finally:
        reset_test_manifest()
        _get_gcs_blob(DEFAULT_REMOTE_PREFIX, TEST_REMOTE_RELATIVE_PATH).delete()


def test_removing_file_works():
    try:
        print(f"Loading manifest: {TEST_MANIFEST_FNAME}")
        manifest = DataManifest(TEST_MANIFEST_FNAME, remote_prefix=DEFAULT_REMOTE_PREFIX)
        print(f"Adding file to manifest.")
        manifest.add_file(
            'eight_As', TEST_DATA_FILE, TEST_LOCAL_RELATIVE_PATH, TEST_REMOTE_RELATIVE_PATH)
        # make sure that the file is in the manifest
        manifest_2 = DataManifest(TEST_MANIFEST_FNAME, remote_prefix=DEFAULT_REMOTE_PREFIX)
        assert 'eight_As' in manifest_2
        print(f"Removing file from manifest.")
        manifest_2.remove_file('eight_As')
        manifest_3 = DataManifest(TEST_MANIFEST_FNAME, remote_prefix=DEFAULT_REMOTE_PREFIX)
        assert 'eight_As' not in manifest_3
    finally:
        reset_test_manifest()
        _get_gcs_blob(DEFAULT_REMOTE_PREFIX, TEST_REMOTE_RELATIVE_PATH).delete()


def test_sync():
    try:
        local_test_data_prefix = tempfile.mkdtemp()
        print(f"Created temp directory '{local_test_data_prefix}'")
        print(f"Loading manifest: {TEST_MANIFEST_FNAME}")
        manifest = DataManifest(TEST_MANIFEST_FNAME, remote_prefix=DEFAULT_REMOTE_PREFIX)
        print(LOCAL_TEST_DATA_PREFIX)
        manifest.sync(LOCAL_TEST_DATA_PREFIX)
        print(f"Synced data to '{LOCAL_TEST_DATA_PREFIX}'")
    finally:
        reset_test_manifest()
        shutil.rmtree(local_test_data_prefix)
        _get_gcs_blob(DEFAULT_REMOTE_PREFIX, TEST_REMOTE_RELATIVE_PATH).delete()


def test_sync_add_sync():
    try:
        local_test_data_prefix = tempfile.mkdtemp()
        print(f"Created temp directory '{local_test_data_prefix}'")
        print(f"Loading manifest: {TEST_MANIFEST_FNAME}")
        manifest = DataManifest(TEST_MANIFEST_FNAME, remote_prefix=DEFAULT_REMOTE_PREFIX)
        print(f"Syncing data to '{LOCAL_TEST_DATA_PREFIX}'")
        manifest.sync(LOCAL_TEST_DATA_PREFIX)
        print(f"Adding file to manifest.")
        manifest.add_file(
            'eight_As', TEST_DATA_FILE, TEST_LOCAL_RELATIVE_PATH, TEST_REMOTE_RELATIVE_PATH)
        print(f"Re-Syncing data to '{LOCAL_TEST_DATA_PREFIX}'")
        manifest.sync(LOCAL_TEST_DATA_PREFIX)
    finally:
        reset_test_manifest()
        shutil.rmtree(local_test_data_prefix)
        _get_gcs_blob(DEFAULT_REMOTE_PREFIX, TEST_REMOTE_RELATIVE_PATH).delete()


def test_verify_success():
    try:
        local_test_data_prefix = tempfile.mkdtemp()
        print(f"Created temp directory '{local_test_data_prefix}'")
        os.mkdir(os.path.join(local_test_data_prefix, "./hg38"))
        shutil.copy(
            TEST_DATA_FILE_2,
            os.path.join(local_test_data_prefix, TEST_2_LOCAL_RELATIVE_PATH)
        )
        print(f"Loading manifest: {TEST_MANIFEST_FNAME}")
        manifest = DataManifest(TEST_MANIFEST_FNAME, remote_prefix=DEFAULT_REMOTE_PREFIX)
        print(local_test_data_prefix)
        manifest.sync(local_test_data_prefix)
        print(f"Synced data to '{local_test_data_prefix}'")
    finally:
        reset_test_manifest()
        print(local_test_data_prefix)
        # shutil.rmtree(local_test_data_prefix)
        _get_gcs_blob(DEFAULT_REMOTE_PREFIX, TEST_REMOTE_RELATIVE_PATH).delete()


def test_verify_size_fail():
    try:
        local_test_data_prefix = tempfile.mkdtemp()
        print(f"Created temp directory '{local_test_data_prefix}'")
        os.mkdir(os.path.join(local_test_data_prefix, "./hg38"))
        shutil.copy(
            # we are copying the wrong file to induce an error
            TEST_DATA_FILE,
            os.path.join(local_test_data_prefix, TEST_2_LOCAL_RELATIVE_PATH)
        )
        print(f"Loading manifest: {TEST_MANIFEST_FNAME}")
        manifest = DataManifest(TEST_MANIFEST_FNAME, remote_prefix=DEFAULT_REMOTE_PREFIX)
        print(local_test_data_prefix)
        manifest.sync(local_test_data_prefix)
        print(f"Synced data to '{local_test_data_prefix}'")
    finally:
        reset_test_manifest()
        print(local_test_data_prefix)
        # shutil.rmtree(local_test_data_prefix)
        _get_gcs_blob(DEFAULT_REMOTE_PREFIX, TEST_REMOTE_RELATIVE_PATH).delete()


def test_verify_md5_fail():
    try:
        local_test_data_prefix = tempfile.mkdtemp()
        print(f"Created temp directory '{local_test_data_prefix}'")
        os.mkdir(os.path.join(local_test_data_prefix, "./hg38"))
        shutil.copy(
            # we are copying the wrong file to induce an error
            TEST_DATA_FILE,
            os.path.join(local_test_data_prefix, TEST_2_LOCAL_RELATIVE_PATH)
        )
        print(f"Loading manifest: {TEST_MANIFEST_FNAME}")
        manifest = DataManifest(TEST_MANIFEST_FNAME, remote_prefix=DEFAULT_REMOTE_PREFIX)
        print(local_test_data_prefix)
        manifest.sync(local_test_data_prefix)
        print(f"Synced data to '{local_test_data_prefix}'")
    finally:
        reset_test_manifest()
        print(local_test_data_prefix)
        # shutil.rmtree(local_test_data_prefix)
        _get_gcs_blob(DEFAULT_REMOTE_PREFIX, TEST_REMOTE_RELATIVE_PATH).delete()

# print("test_add_remote_missing_file")
# test_add_remote_missing_file()
# print("test_add_remote_present_file")
# test_add_remote_present_file()
# print("test_add_remote_mismatch_file")
# test_add_remote_mismatch_file()
# print("test_locking_works")
# test_locking_works()
# print("test_adding_two_files_works")
# test_adding_two_files_works()
# print("test_removing_file_works")
# test_removing_file_works()
# print("test_sync")
# test_sync()
# print("test_sync_add_sync")
# test_sync_add_sync()
# print("test_add_duplicate_key")
# test_add_duplicate_key()
# print("test_verify_success")
# test_verify_success()
# print("test_verify_fail")
# test_verify_fail()


def parse_args():
    # add data file
    # remove data file
    # replace data file
    pass


def main(fname):
    with open(fname) as ifp:
        for i, line in enumerate(ifp):
            if i == 0:
                print(line.strip("\n"))
                continue
            data = line.strip("\n").split("\t")
            # add space for the md5sum, filesize, and last update time
            data.insert(-1, None)
            data.insert(-1, None)
            data.insert(-1, None)

            data[-1] += "#old path: {}".format(data[1])
            # update to use the new path
            local_filename = data[2].replace("gs://balrog/", "/srv/")
            data[1] = local_filename
            # calculate the md5sum
            md5sum = subprocess.run(
                ["md5sum", data[1]],
                stdout=subprocess.PIPE
            ).stdout.split()[0].decode('UTF8')
            data[-4] = md5sum
            file_size = str(os.path.getsize(local_filename))
            data[-3] = file_size
            last_mtime = str(int(os.path.getmtime(local_filename)))
            data[-2] = last_mtime
            print("\t".join(data))
    return

# if __name__ == '__main__':
#     main(sys.argv[1])



import os
import sys
import logging

logging.basicConfig(
    stream=sys.stderr, format='%(levelname)s\t%(asctime)-15s\t%(message)s', level=logging.DEBUG
)


def _get_version():
    """Load version from VERSION file.

    VERSION is stored in the base of the repo, but is copied into the module
    directory when building the distribution. Therefore we need to look for the
    VERSION file in two locations: one directory up and in the current directory.
    """

    def _get_version_from_file_or_none(version_fname):
        try:
            with open(version_fname) as fp:
                return fp.read().strip()
        except FileNotFoundError:
            return None

    develop_version = _get_version_from_file_or_none(
        os.path.join(os.path.dirname(__file__), "../VERSION"))
    packaged_version = _get_version_from_file_or_none(
        os.path.join(os.path.dirname(__file__), "./VERSION"))
    if develop_version is not None:
        assert packaged_version is None or packaged_version == develop_version, \
            "VERSION files must match if they both exist"
        return develop_version
    elif packaged_version is not None:
        assert develop_version is None or packaged_version == develop_version, \
            "VERSION files must match if they both exist"
        return packaged_version
    else:
        raise RuntimeError("Can not find VERSION file.")


__version__ = _get_version()



import os
import re
import contextlib
import subprocess
import logging

import conda_build.api
from conda_build.config import Config as CondaBuildConfig


logger = logging.getLogger(__file__)  # noqa: invalid-name


class YamlNotFoundError(Exception):
    pass


def norm_abs_join_path(*paths):
    return os.path.normpath(os.path.abspath(os.path.join(*paths)))


def get_yaml_path(repo_path):
    yaml_fpath = norm_abs_join_path(repo_path, "conda-build/meta.yaml")
    if not os.path.exists(yaml_fpath):
        raise YamlNotFoundError(f"Could not find a meta.yaml file at '{yaml_fpath}'")
    else:
        return yaml_fpath


def run_and_log(cmd, input=None):
    logger.info(f"Running '{cmd}'")

    proc = subprocess.run(
        cmd, shell=True, input=input, stdout=subprocess.PIPE, stderr=subprocess.PIPE
    )

    if proc.stderr:
        logger.info(f"Ran '{cmd}'\nSTDERR:\n{proc.stderr.decode().strip()}")
    if proc.stdout:
        logger.info(f"Ran '{cmd}'\nSTDOUT:\n{proc.stdout.decode().strip()}")

    # raise an excpetion if the return code was non-zero
    proc.check_returncode()

    return proc


@contextlib.contextmanager
def change_directory(path):
    """A context manager which changes the working directory to the given
    path, and then changes it back to its previous value on exit.

    """
    prev_cwd = os.getcwd()
    os.chdir(path)
    yield
    os.chdir(prev_cwd)


def get_git_repo_name(path):
    with change_directory(path):
        proc = subprocess.run(
            "git config --get remote.origin.url",
            shell=True, stdout=subprocess.PIPE, check=True
        )
        res = proc.stdout.decode().strip()
        # extract the basename, and then strip '.git' off of the end
        return os.path.basename(res)[:-4]


def build_package_from_meta_yaml(path, version, skip_existing=False):
    # Set the environment variable VERSION so that
    # the jinja2 templating works for the conda-build
    local_env = os.environ
    local_env['VERSION'] = version

    # build the package
    yaml_fpath = get_yaml_path(path)
    output_file_paths = conda_build.api.build(
        [yaml_fpath, ],
        skip_existing=skip_existing,
        config=CondaBuildConfig(anaconda_upload=False, quiet=True)
    )
    if len(output_file_paths) == 0:
        raise RuntimeError('No package was built.')
    assert len(output_file_paths) == 1, \
        "multiple file paths in conda build: {}".format(str(output_file_paths))
    return output_file_paths[0]


def build_package_using_distutils(path):
    proc = subprocess.run(
        "python setup.py bdist_conda",
        shell=True, check=True, stdout=subprocess.PIPE
    )
    upload_pat = "# \$ anaconda upload (\S+)$"
    build_output = proc.stdout.decode()
    for line in build_output.splitlines():
        res = re.findall(upload_pat, line)
        if res:
            assert len(res) == 1
            return res[0]

    logger.debug('STDOUT: {}'.format(build_output))
    raise ValueError('Could not extract package file from stdout.')


def build_package(path, version, skip_existing=False):
    try:
        yaml_path = get_yaml_path(path)
    except YamlNotFoundError:
        yaml_path = None

    # if we can't find the yaml file, install using disttools bdist_conda
    if yaml_path is None:
        return build_package_using_distutils(path)
    else:
        return build_package_from_meta_yaml(path=path, version=version, skip_existing=skip_existing)



import os
import sys
import subprocess

from freenome_build.util import build_package
from freenome_build import version_utils

LOCAL_CONDA_BUILD_SCRIPT = os.path.abspath('scripts/conda_build.sh')


def build_and_upload_package_from_repo(path='./', upload=True, skip_existing=False, repo_name=None):
    """

    Args:
        path (str): library path to install (default './')
        upload (bool): upload to freenome conda channel (default True)
        skip_existing (bool): do not build if existing build in local conda install (default False)
        repo_name (str): repo name to build (optional)

    Returns:
        None

    """
    version = version_utils.version(path)
    output_file_path = build_package(path, version, skip_existing=skip_existing)

    if upload:
        upload_cmd = ['anaconda', '-t', os.environ['ANACONDA_TOKEN'],
                      'upload', '--force', '-u', 'freenome', output_file_path]

        subprocess.check_call(upload_cmd, stdout=sys.stdout, stderr=sys.stderr)


def deploy_main(args):
    return build_and_upload_package_from_repo(
        path=args.path,
        upload=args.upload,
        skip_existing=args.skip_existing
    )


def add_deploy_subparser(subparsers):
    # deploy parser
    deploy_subparser = subparsers.add_parser('deploy', help='deploy a package')
    deploy_subparser.add_argument(
        '-u', '--upload', action='store_true', default=False, dest='upload')
    deploy_subparser.add_argument(
        '-p', '--path', action='store', default='./', dest='path')
    deploy_subparser.add_argument(
        '--skip', action='store_true', default=False, dest='skip_existing')



import os
import time
import subprocess
import logging
import psycopg2

from freenome_build.util import norm_abs_join_path, change_directory, get_git_repo_name, run_and_log

logger = logging.getLogger(__file__)  # noqa: invalid-name

DEFAULT_TEST_DB_PORT = 3333
DEFAULT_TEST_DB_HOST = 'localhost'

# the maximum amount of time in seconds to wait for the DB to come up before raising an error
MAX_DB_WAIT_TIME = 10


class ContainerDoesNotExistError(Exception):
    pass


def _execute_sql_script(sql_script_path, dbuser, dbname, host, port):
    pass


def _setup_db(image_name, repo_path, host, port):
    # check if 'setup' exists in repo_path/database/
    repo_setup_sql_path = norm_abs_join_path(repo_path, "./database/setup.sql")
    if os.path.exists(repo_setup_sql_path):
        with open(repo_setup_sql_path) as ifp:
            setup_sql = ifp.read()
    # if this doesn't exist, revert to the default
    else:
        setup_sql_template_path = norm_abs_join_path(
            os.path.dirname(__file__), "./database_template/scripts/setup.sql")
        logger.debug(f"The repo at '{repo_path}' does not contain './database/setup.sql'."
                     f"\nDefaulting to {setup_sql_template_path} with "
                     f"USER='{image_name}', DATABASE='{image_name}'")
        # get the new database password from the environment
        if 'PGPASSWORD' not in os.environ:
            raise ValueError("PGPASSWORD must be in the environment to setup a new database.")
        with open(setup_sql_template_path) as ifp:
            setup_sql = ifp.read().format(
                PGUSER=image_name, PGDATABASE=image_name, PGPASSWORD=os.environ['PGPASSWORD']
            )

    # execute the setup.sql script
    run_and_log(f"psql -h {host} -p {port} -U postgres", input=setup_sql.encode())

    return


def _run_migrations(repo_path, host, port, dbname, dbuser):
    # check if 'migrate' exists in repo_path/database/
    repo_migrate_path = norm_abs_join_path(repo_path, "./database/migrate")
    if os.path.exists(repo_migrate_path):
        # TODO -- add support for running this script
        raise NotImplementedError("We have not implemented running migrate from the repo.")

    logger.info(f"The repo at '{repo_path}' does not contain './database/migrate' script"
                "\nDefaulting to running sqitch migrations in './database/sqitch'")
    sqitch_path = norm_abs_join_path(repo_path, "./database/sqitch")
    if not os.path.exists(sqitch_path):
        raise RuntimeError(
            f"Sqitch migration files must exist at '{sqitch_path}' "
            "if a migration script is not provided.")

    with change_directory(sqitch_path):
        try:
            run_and_log(
                f"sqitch --engine pg deploy db:pg://postgres@{host}:{port}/{dbname}")
        except subprocess.CalledProcessError as inst:
            # we don't care if there's nothing to deploy
            if inst.stderr.decode().strip() == 'Nothing to deploy (empty plan)':
                pass
            else:
                raise
    return


def _insert_test_data(repo_path, host, port, dbuser, dbname):
    # check if 'migrate' exists in repo_path/database/
    repo_insert_test_data_path = norm_abs_join_path(
        repo_path, "./database/insert_test_data")
    if os.path.exists(repo_insert_test_data_path):
        # TODO -- add support for running this script
        raise NotImplementedError("We have not implemented running insert_test_data from the repo.")
    else:
        logger.debug(
            f"The repo at '{repo_path}' does not contain './database/insert_test_data' script")

    repo_insert_test_data_sql_path = norm_abs_join_path(
        repo_path, "./database/insert_test_data.sql")
    if os.path.exists(repo_insert_test_data_sql_path):
        logger.info(f"Inserting data in '{repo_insert_test_data_sql_path}'.")
        with open(repo_insert_test_data_sql_path) as ifp:
            run_and_log(
                f"psql -h {host} -p {port} -U {dbuser} {dbname}",
                input=ifp.read().encode()
            )
        return
    else:
        logger.info(
            f"The repo at '{repo_path}' does not contain './database/insert_test_data.sql' script")

    raise ValueError(f"'{repo_path}' does not contain an insert test data script or sql file.")


def _wait_for_db_cluster_to_start(host, port, max_wait_time=MAX_DB_WAIT_TIME, recheck_interval=0.2):
    conn_str = f"dbname=postgres user=postgres host={host} port={port}"
    for _ in range(int(max_wait_time/recheck_interval)+1):
        try:
            with psycopg2.connect(conn_str) as _: # noqa
                # we just want to see if we can connect, so if we do then connect
                logger.debug(f"Database at '{conn_str}' is up!")
                return
        except psycopg2.OperationalError:
            logger.debug(f"DB cluster at '{conn_str}' is not yet up.")
            time.sleep(recheck_interval)
            continue

    raise RuntimeError(f"Aborting because the DB did not start within {MAX_DB_WAIT_TIME} seconds.")


def stop_test_database(project_name, host=DEFAULT_TEST_DB_HOST, port=DEFAULT_TEST_DB_PORT):
    if host != 'localhost':
        raise NotImplementedError('Non localhost test databases are not supported.')
    image_name = f"{project_name}_{port}"
    cmd = f"docker kill {image_name}"
    try:
        run_and_log(cmd)
    except subprocess.CalledProcessError as inst:
        # if this is an error because the container already exists, then raise
        # a custom error type
        pat = (f"Error response from daemon: Cannot kill container:"
               f" {image_name}: No such container: {image_name}")
        if inst.stderr.decode().strip() == pat:
            raise ContainerDoesNotExistError(inst)
        # otherwise just propogate the error
        else:
            raise

    cmd = f"docker rm -f {image_name}"
    run_and_log(cmd)


def start_test_database(
        repo_path, project_name, host=DEFAULT_TEST_DB_HOST, port=DEFAULT_TEST_DB_PORT):
    """Start a test database in a docker container.

    This starts a new test database in a docker container. This function:
    1) builds the postgres server docker image
    2) starts the docker container on port 'port'
    3) waits for the database cluster to start
    4) runs the DB setup script
       - create a new db user $dbname
       - create a new db $dbname owned by $dbname
    5) runs the database migrations
    """
    # TODO (nb): add a check to ensure that 'port' is free
    # host is always localhost because we are running it in a local Docker container
    if host != 'localhost':
        raise NotImplementedError('Non localhost test databases are not supported.')

    # set the path to the Postgres Dockerfile
    docker_file_path = norm_abs_join_path(repo_path, "./database/Dockerfile")
    # if the repo doesn't have a Dockerfile in the database sub-directory, then
    # default to the template Dockerfile
    if not os.path.exists(docker_file_path):
        docker_file_path = norm_abs_join_path(
            os.path.dirname(__file__), "./database_template/Dockerfile")
        logger.info(f"Setting DB docker file path to '{docker_file_path}'")

    docker_file_dir = os.path.dirname(docker_file_path)

    # build
    cmd = f"docker build --rm -t {project_name}:latest {docker_file_dir}"
    run_and_log(cmd)

    # starting-db
    cmd = f"docker run -d -p {port}:5432 --name {project_name}_{port} {project_name}:latest"
    run_and_log(cmd)
    # the database cluster needs some time to start, so try to connect periodically until we can
    _wait_for_db_cluster_to_start(host, port)

    # setup-db
    # we need to connect to the 'postgres' database to create a new database
    _setup_db(project_name, repo_path, host, port)

    # run-migrations
    _run_migrations(
        repo_path=repo_path, host=host, port=port, dbname=project_name, dbuser=project_name)

    # insert test data
    _insert_test_data(
        repo_path=repo_path, host=host, port=port, dbname=project_name, dbuser=project_name)

    # log the connection command
    connection_cmd = f"psql -h {host} -p {port} -U {project_name} {project_name}"
    logger.info(f"Database is up! You can connect by running:\n{connection_cmd}")


def start_test_database_main(args):
    if args.force:
        # if the container does not exist then we don't need to stop anything
        try:
            stop_test_database(args.project_name, args.host, args.port)
        except ContainerDoesNotExistError:
            pass
    start_test_database(args.path, args.project_name, args.host, args.port)


def stop_test_database_main(args):
    stop_test_database(args.project_name, args.host, args.port)


def connect_to_test_db_main(project_name, host, port):
    connection_cmd = f"psql -h {host} -p {port} -U {project_name} {project_name}"
    logger.debug(f"Running '{connection_cmd}'")
    proc = subprocess.Popen(connection_cmd, shell=True)
    proc.wait()


def add_db_subparser(subparsers):
    database_parser = subparsers.add_parser('test-db', help='manage the test database')
    database_parser.required = True
    database_parser.add_argument('--path', default='.')
    database_parser.add_argument(
        '--port', default=DEFAULT_TEST_DB_PORT,
        help='Port on which to start the test db. Default: %(default)s'
    )
    database_parser.add_argument(
        '--host', default=DEFAULT_TEST_DB_HOST,
        help='Teste DB host. Default: %(default)s'
    )
    database_parser.add_argument(
        '--project-name',
        default=None,
        help='The name of the project.\n'
             'This is assumed to be the name of the database and the database owner.\n'
             'Default: The name of the git repo at $PATH with - replaced with _'
    )

    # add the subparsers
    database_subparsers = database_parser.add_subparsers(dest='test_db_command')
    database_subparsers.required = True

    # Start a test DB
    start_test_db_parser = database_subparsers.add_parser('start', help='start a test database')
    start_test_db_parser.add_argument(
        '--force', '-f', action='store_true', default=False,
        help='Start the test database even if it means stopping an already existing '
             'docker container of the same name.'
    )

    # Stop the test db
    database_subparsers.add_parser('stop', help='stop the test database')

    # Connect to a test db
    database_subparsers.add_parser('connect', help='Connect to the test database.')


def db_main(args):
    # normalize the path
    args.path = norm_abs_join_path(args.path)

    # set the project name to the name of the git repo at args.path
    if args.project_name is None:
        args.project_name = get_git_repo_name(args.path).replace('-', '_')
        logger.info(f"Setting project name to '{args.project_name}'")

    if args.test_db_command == 'connect':
        connect_to_test_db_main(args.project_name, args.host, args.port)
    elif args.test_db_command == 'start':
        start_test_database_main(args)
    elif args.test_db_command == 'stop':
        stop_test_database_main(args)
    else:
        raise ValueError(f"Unrecognized DB subcommand '{args.test_db_command}'")



import os
import logging
import yaml

from freenome_build.util import build_package, run_and_log, norm_abs_join_path, change_directory
from freenome_build.github import repo_name
from freenome_build import version_utils

logger = logging.getLogger(__file__)  # noqa: invalid-name


def get_package_name_from_meta_yaml(path):
    with open(norm_abs_join_path(path, './conda-build/meta.yaml')) as ifp:
        data_template = ifp.read()
        # replace the VERSION template with 0, because we don't actually care
        # about the version to get the package name, but we need to yaml to be parseable
        data = yaml.load(data_template.replace("{{ VERSION }}", "0"))
        return data['package']['name']


def setup_development_environment(path):
    version = version_utils.version(path)
    logging.debug('version: %s', version)

    # get package name
    try:
        package_name = get_package_name_from_meta_yaml(path)
    except FileNotFoundError:
        with change_directory(path):
            package_name = repo_name().replace("_", "-")

    logging.debug('package name: %s', package_name)

    # build the package.
    # ( we need to do this to install the dependencies -- which is super hacky but required
    #   because of the jinja templating -- see https://github.com/conda/conda/issues/5126   )
    output_file_path = build_package(path, version=version, skip_existing=False)
    logger.debug('output build to %s', output_file_path)

    # conda install this package, which installs all of the dependencies
    logger.debug("installing package at '%s'", output_file_path)
    # extract the local path from the return output_file_path, which is of the form:
    # /tmp/nboley/conda/linux-64/balrog-10-0.tar.bz2
    local_channel = "file://" + os.path.split(os.path.split(output_file_path)[0])[0]
    run_and_log(f"conda install {package_name}=={version} --only-deps --yes -c {local_channel}")
    # python setup.py develop $PATH
    run_and_log("python {} develop".format(norm_abs_join_path(path, "./setup.py")))


def add_develop_subparser(subparsers):
    develop_subparser = subparsers.add_parser(
        'develop', help='initialize a development environment')
    develop_subparser.required = True
    develop_subparser.add_argument('path', default='.')


def develop_main(args):
    setup_development_environment(args.path)



import logging
import os
import re

from freenome_build import github


def version(path, repo_name=None):
    """Return version id

    Args:
        repo_name: repo name (optional)

    Returns:
        (str): version id
    """
    if repo_name is None:
        repo_name = github.repo_name()

    version_from_init = get_version_from_init(path, repo_name)
    version_from_version_file = get_version_from_version_file(path)

    assert version_from_init or version_from_version_file, "Version file cannot be found"
    assert (version_from_init is not None) ^ (version_from_version_file is not None), "Multiple version files found"

    if version_from_init:
        return version_from_init
    elif version_from_version_file:
        return version_from_version_file
    else:
        raise FileNotFoundError("Version file cannot be found.")


def get_version_from_init(path, repo_name=None):
    """get version from init in repo with directory structure:

    lib_name
    lib_name/lib_name
    lib_name/lib_name/__init__.py

    Args:
        repo_name: repo name

    Returns:
        (str): version id

    """
    if repo_name is None:
        return None

    repo_init_filepath = os.path.abspath(os.path.join(path, './{}/__init__.py'.format(repo_name)))
    logging.debug('repo_init_filepath: {}'.format(repo_init_filepath))

    if os.path.exists(repo_init_filepath):
        with open(repo_init_filepath, 'r') as fp:
            # returns string following "__version__ = "
            search = re.search(r'^__version__\s*=\s*[\'"]([^\'"]*)[\'"]',
                               fp.read(), re.MULTILINE)
            version = search.group(1) if search is not None else None

        return version


def get_version_from_version_file(path):
    """get version from a VERSION file in the repo

    {PATH}/VERSION

    Returns:
        (str): version id

    """
    version_filepath = os.path.abspath(os.path.join(path, './VERSION'))
    logging.debug('version_filepath: {}'.format(version_filepath))

    if os.path.exists(version_filepath):
        with open(version_filepath, 'r') as fp:
            version = fp.read().strip()
        logging.debug("Found version: '{}'".format(version))
        return version



import os
import setuptools

try:
    from pip.download import PipSession
    from pip.req import parse_requirements
except ImportError:
    from pip._internal.download import PipSession
    from pip._internal.req import parse_requirements


VERSION_FILEPATH = os.path.abspath(os.path.join(os.path.dirname(__file__), './VERSION'))


def get_version():
    if not os.path.exists(VERSION_FILEPATH):
        raise RuntimeError(f"Can not find version file at '{VERSION_FILEPATH}'")

    with open(VERSION_FILEPATH, 'r') as fp:
        return fp.read().strip()


def main():
    version = get_version()
    install_reqs = parse_requirements("update_requirements.txt", session=PipSession())
    reqs = [str(ir.req) for ir in install_reqs if ir.req]

    setuptools.setup(
        name='freenome-build',
        version=version,
        description='Freenome build',
        url='https://github.com/freenome/freenome-build',
        author='Nathan Boley',
        author_email="nathan.boley@freenome.com",

        install_requires=reqs + ["pytest"],
        package_data={
            '': ['./VERSION'],
            'freenome_build': [
                './database_template/*',
                './database_template/*/*',
            ]
        },
        packages=setuptools.find_packages(),
        scripts=['bin/freenome-build']
    )


if __name__ == '__main__':
    main()



import os
import subprocess


def repo_name():
    """Return github repository name of current directory

    Returns:
        (str): repo name
    """
    return os.path.basename(
        subprocess.run(
            'git rev-parse --show-toplevel',
            shell=True,
            stdout=subprocess.PIPE
        ).stdout.strip().decode('utf8')).lower().replace('-', '_')


def environment_name():
    """Returns local environment name

    Returns:
        (str): environment name
    """
    return subprocess.run(
        'git describe --all HEAD^',
        shell=True,
        stdout=subprocess.PIPE
    ).stdout.strip().decode('utf8').replace('/', '__')
